baseVersion = '8.0.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '8.0.0-d2ae0ea9ab49f238670dbf6aafd20de7afdd8514'
cacheTag = 'cache-local-py3.13'
mainCacheTag = 'cache-master-py3.13'
distVersion = '8.0.0'
exactPython = 'python3.13'
python = 'python3.13'
dockerTag = '8.0.0-d2ae0ea9ab49f238670dbf6aafd20de7afdd8514-py3.13'
currentCommit = 'd2ae0ea9ab49f238670dbf6aafd20de7afdd8514'
dockerRegistry = 'quay.io/stxue'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20250110105449'
